#!/usr/bin/env python
from pymap3d.vincenty import vreckon
from argparse import ArgumentParser


def main():
    p = ArgumentParser(description='Python port of vreckon.m')
    p.add_argument('lat', help='latitude WGS-84 [degrees]', type=float)
    p.add_argument('lon', help='longitude WGS-84 [degrees]', type=float)
    p.add_argument('range', help='range from start point [meters]', type=float)
    p.add_argument('azimuth', help='azimuth to start [deg.]', type=float)
    P = p.parse_args()

    lat2, lon2, a21 = vreckon(P.lat, P.lon, P.range, P.azimuth)
    print('new (lat, lon)  ({}, {}) '.format(lat2, lon2))
    print('az back to start:', a21)


if __name__ == '__main__':  # pragma: no cover
    main()



import pytest
from pytest import approx
import pymap3d as pm

lat, lon = (65, -148)
lla0 = (42, -82, 200)
azi, eli = (180.1, 80)
t0 = '2014-04-06T08:00:00'
ra, dec = (166.5032081149338, 55.000011165405752)


def test_azel2radec():
    R, D = pm.azel2radec(azi, eli, lat, lon, t0)
    assert R == approx(ra, rel=0.01)
    assert D == approx(dec, rel=0.01)

    Rv, Dv = pm.azel2radec(azi, eli, lat, lon, t0, usevallado=True)
    assert Rv == approx(ra)
    assert Dv == approx(dec)


def test_radec2azel():
    azapy, elapy = pm.radec2azel(ra, dec, lat, lon, t0)
    assert azapy == approx(azi, rel=0.01, abs=0.1)
    assert elapy == approx(eli, rel=0.01, abs=0.1)

    azvallado, elvallado = pm.radec2azel(ra, dec, lat, lon, t0, usevallado=True)
    assert azvallado == approx(azi, rel=1e-2)
    assert elvallado == approx(eli, rel=1e-2)


if __name__ == '__main__':
    pytest.main(['-xrsv', __file__])



"""
Azimuth / elevation <==> Right ascension, declination
"""
from typing import Tuple
from datetime import datetime
import numpy as np
from .vallado import azel2radec as vazel2radec, radec2azel as vradec2azel
from .timeconv import str2dt  # astropy can't handle xarray times (yet)
try:
    from astropy.time import Time
    from astropy import units as u
    from astropy.coordinates import Angle, SkyCoord, EarthLocation, AltAz, ICRS
except ImportError:
    Time = None


def azel2radec(az_deg: float, el_deg: float,
               lat_deg: float, lon_deg: float,
               time: datetime, usevallado: bool=False) -> Tuple[float, float]:
    """
    viewing angle (az, el) to sky coordinates (ra, dec)

    inputs
    ------
    azimuth: degrees clockwize from North
    elevation: degrees above horizon (neglecting aberration)
    observer latitude [-90, 90], longitude [-180, 180] (degrees)
    time: datetime of observation

    Outputs
    -------
    ecliptic right ascension, declination (degrees)
    """

    if usevallado or Time is None:  # non-AstroPy method, less accurate
        return vazel2radec(az_deg, el_deg, lat_deg, lon_deg, time)

    obs = EarthLocation(lat=lat_deg * u.deg, lon=lon_deg * u.deg)

    direc = AltAz(location=obs, obstime=Time(str2dt(time)),
                  az=az_deg * u.deg, alt=el_deg * u.deg)

    sky = SkyCoord(direc.transform_to(ICRS()))

    return sky.ra.deg, sky.dec.deg


def radec2azel(ra_deg: float, dec_deg: float,
               lat_deg: float, lon_deg: float,
               time: datetime, usevallado: bool=False) -> Tuple[float, float]:
    """
    sky coordinates (ra, dec) to viewing angle (az, el)

    inputs
    ------
    ecliptic right ascension, declination (degrees)
    observer latitude [-90, 90], longitude [-180, 180] (degrees)
    time: datetime of observation

    Outputs
    -------
    azimuth: degrees clockwize from North
    elevation: degrees above horizon (neglecting aberration)
    """

    if usevallado or Time is None:
        return vradec2azel(ra_deg, dec_deg, lat_deg, lon_deg, time)
# %% input trapping
    lat = np.atleast_1d(lat_deg)
    lon = np.atleast_1d(lon_deg)
    ra = np.atleast_1d(ra_deg)
    dec = np.atleast_1d(dec_deg)

    obs = EarthLocation(lat=lat * u.deg,
                        lon=lon * u.deg)

    points = SkyCoord(Angle(ra, unit=u.deg),
                      Angle(dec, unit=u.deg),
                      equinox='J2000.0')

    altaz = points.transform_to(AltAz(location=obs, obstime=Time(str2dt(time))))

    return altaz.az.degree, altaz.alt.degree



import numpy as np
from .ecef import Ellipsoid


def isometric(lat: float, ell: Ellipsoid=None, deg: bool=True):
    """
     computes isometric latitude of a point on an ellipsoid

      input
      -----
      lat    latitude (degrees/radians)
      ell    reference ellipsoid
      deg    degrees input/output  (False: radians in/out)

      output
      ------
      isolat  isometric latiude (degrees/radians)

     notes:
     isometric latitude is an auxiliary latitude proportional to the spacing
     of parallels of latitude on an ellipsoidal mercator projection.

     Based on Deakin, R.E., 2010, 'The Loxodrome on an Ellipsoid', Lecture Notes,
     School of Mathematical and Geospatial Sciences, RMIT University,
     January 2010
    """

    if ell is None:
        ell = Ellipsoid()

    f = ell.f  # flattening of ellipsoid

    if deg is True:
        lat = np.deg2rad(lat)

    e2 = f * (2 - f)  # eccentricity-squared
    e = np.sqrt(e2)  # eccentricity of ellipsoid

    x = e * np.sin(lat)
    y = (1 - x) / (1 + x)
    z = np.pi / 4 + lat / 2

#   calculate the isometric latitude
    isolat = np.log(np.tan(z) * (y**(e / 2)))

    if deg is True:
        isolat = np.degrees(isolat)

    return isolat


def meridian_dist(lat: float, ell: Ellipsoid=None, deg: bool=True):
    """
    computes the meridian distance on an ellipsoid *from the equator* to a latitude.

     input
     -----
     lat    latitude (degrees/radians)
     ell    reference ellipsoid
     deg    degrees input/output  (False: radians in/out)

     output
     ------
     mdist  meridian distance (degrees/radians)

    notes:
    Formula given Baeschlin, C.F., 1948,
    "Lehrbuch Der Geodasie", Orell Fussli Verlag, Zurich, pp.47-50.

    Based on Deakin, R.E., 2010, 'The Loxodrome on an Ellipsoid', Lecture Notes,
    School of Mathematical and Geospatial Sciences, RMIT University, January 2010
    """

    if deg is True:
        lat = np.radians(lat)

    #   set ellipsoid parameters
    if ell is None:
        ell = Ellipsoid()

    a = ell.a
    f = ell.f  # flattening of ellipsoid

    e2 = f * (2 - f)  # eccentricity-squared

    # powers of eccentricity
    e4 = e2 * e2
    e6 = e4 * e2
    e8 = e6 * e2
    e10 = e8 * e2

    # coefficients of series expansion for meridian distance
    A = 1 + (3 / 4) * e2 + (45 / 64) * e4 + (175 / 256) * e6 + (11025 / 16384) * e8 + (43659 / 65536) * e10
    B = (3 / 4) * e2 + (15 / 16) * e4 + (525 / 512) * e6 + (2205 / 2048) * e8 + (72765 / 65536) * e10
    C = (15 / 64) * e4 + (105 / 256) * e6 + (2205 / 4096) * e8 + (10395 / 16384) * e10
    D = (35 / 512) * e6 + (315 / 2048) * e8 + (31185 / 131072) * e10
    E = (315 / 16384) * e8 + (3465 / 65536) * e10
    F = (693 / 131072) * e10

    term1 = A * lat
    term2 = (B / 2) * np.sin(2 * lat)
    term3 = (C / 4) * np.sin(4 * lat)
    term4 = (D / 6) * np.sin(6 * lat)
    term5 = (E / 8) * np.sin(8 * lat)
    term6 = (F / 10) * np.sin(10 * lat)

    mdist = a * (1 - e2) * (term1 - term2 + term3 - term4 + term5 - term6)

    return mdist


def loxodrome_inverse(lat1: float, lon1: float, lat2: float, lon2: float,
                      ell: Ellipsoid=None, deg: bool=True):
    """
    computes the arc length and azimuth of the loxodrome
    between two points on the surface of the reference ellipsoid

    input
    -----
    lat1
     GEODETIC latitude of first point (degrees/radians)

    lon1
     longitude of first point (degrees/radians)

    lat2, lon2
     second point (degrees/radians)

    ell    reference ellipsoid
    deg    degrees input/output  (False: radians in/out)

    output
    ------
    lox_s  distance along loxodrome
    az12   azimuth of loxodrome (degrees/radians)

    Based on Deakin, R.E., 2010, 'The Loxodrome on an Ellipsoid', Lecture Notes,
    School of Mathematical and Geospatial Sciences, RMIT University, January 2010

    [1] Bowring, B.R., 1985, 'The geometry of the loxodrome on the
    ellipsoid', The Canadian Surveyor, Vol. 39, No. 3, Autumn 1985,
    pp.223-230.
    [2] Snyder, J.P., 1987, Map Projections-A Working Manual. U.S.
    Geological Survey Professional Paper 1395. Washington, DC: U.S.
    Government Printing Office, pp.15-16 and pp. 44-45.
    [3] Thomas, P.D., 1952, Conformal Projections in Geodesy and
    Cartography, Special Publication No. 251, Coast and Geodetic
    Survey, U.S. Department of Commerce, Washington, DC: U.S.
    Government Printing Office, p. 66.
    """

    #   set ellipsoid parameters
    if ell is None:
        ell = Ellipsoid()

    if deg is True:
        lat1, lon1, lat2, lon2 = np.radians([lat1, lon1, lat2, lon2])

    # compute isometric latitude of P1 and P2
    isolat1 = isometric(lat1, deg=False, ell=ell)
    isolat2 = isometric(lat2, deg=False, ell=ell)

    # compute changes in isometric latitude and longitude between points
    disolat = isolat2 - isolat1
    dlon = lon2 - lon1

    # compute azimuth
    az12 = np.arctan2(dlon, disolat)

    # compute distance along loxodromic curve
    m1 = meridian_dist(lat1, deg=False, ell=ell)
    m2 = meridian_dist(lat2, deg=False, ell=ell)
    dm = m2 - m1
    lox_s = dm / np.cos(az12)

    if deg is True:
        az12 = np.degrees(az12) % 360.

    return lox_s, az12



#!/usr/bin/env python
from pymap3d.vincenty import vdist
from argparse import ArgumentParser


def main():
    p = ArgumentParser(description='vdist distance between WGS-84 coordinates')
    p.add_argument('lat1', help='latitude1 WGS-84 [degrees]', type=float)
    p.add_argument('lon1', help='longitude1 WGS-84 [degrees]', type=float)
    p.add_argument('lat2', help='latitude2 WGS-84 [degrees]', type=float)
    p.add_argument('lon2', help='longitude2 WGS-84 [degrees]', type=float)
    P = p.parse_args()

    dist_m = vdist(P.lat1, P.lon1, P.lat2, P.lon2)
    print('distance between WGS-84 points: {} m'.format(dist_m))


if __name__ == '__main__':  # pragma: no cover
    main()



#!/usr/bin/env python
"""
Example Kitt Peak

./demo_azel2radec.py 264.9183 37.911388 31.9583 -111.597 2014-12-25T22:00:00MST
"""
from pymap3d import azel2radec
from argparse import ArgumentParser


def main():
    p = ArgumentParser(description="convert azimuth and elevation to "
                       "right ascension and declination")
    p.add_argument('azimuth', help='azimuth [deg]', type=float)
    p.add_argument('elevation', help='elevation [deg]', type=float)
    p.add_argument('lat', help='WGS84 obs. lat [deg]', type=float)
    p.add_argument('lon', help='WGS84 obs. lon [deg]', type=float)
    p.add_argument('time', help='obs. time YYYY-mm-ddTHH:MM:SSZ')
    P = p.parse_args()

    ra, dec = azel2radec(P.azimuth, P.elevation, P.lat, P.lon, P.time)

    print('ra [deg] ', ra, ' dec [deg] ', dec)


if __name__ == "__main__":  # pragma: no cover
    main()



from numpy import radians, sin, cos, tan, allclose, hypot, degrees, arctan2, sqrt, pi
import numpy as np
from copy import deepcopy
from typing import Tuple
from datetime import datetime

from .eci import eci2ecef


class Ellipsoid:
    """
    generate reference ellipsoid parameters

    https://en.wikibooks.org/wiki/PROJ.4#Spheroid
    """

    def __init__(self, model: str='wgs84') -> None:
        if model == 'wgs84':
            """https://en.wikipedia.org/wiki/World_Geodetic_System#WGS84"""
            self.a = 6378137.  # semi-major axis [m]
            self.f = 1 / 298.2572235630  # flattening
            self.b = self.a * (1 - self.f)  # semi-minor axis
        elif model == 'grs80':
            """https://en.wikipedia.org/wiki/GRS_80"""
            self.a = 6378137.  # semi-major axis [m]
            self.f = 1 / 298.257222100882711243  # flattening
            self.b = self.a * (1 - self.f)  # semi-minor axis
        elif model == 'clrk66':  # Clarke 1866
            self.a = 6378206.4  # semi-major axis [m]
            self.b = 6356583.8  # semi-minor axis
            self.f = -(self.b / self.a - 1)
        elif model == 'mars':  # https://tharsis.gsfc.nasa.gov/geodesy.html
            self.a = 3396900
            self.b = 3376097.80585952
            self.f = 1 / 163.295274386012
        elif model == 'moon':
            self.a = 1738000.
            self.b = 1738000.
            self.f = 0.
        elif model == 'venus':
            self.a = 6051000.
            self.b = 6051000.
            self.f = 0.
        else:
            raise NotImplementedError('{} model not implemented, let us know and we will add it (or make a pull request)'.format(model))


def get_radius_normal(lat_radians: float, ell: Ellipsoid=None) -> float:
    """ Compute normal radius of planetary body"""
    if ell is None:
        ell = Ellipsoid()

    a = ell.a
    b = ell.b

    return a**2 / sqrt(a**2 * cos(lat_radians)**2 + b**2 * sin(lat_radians)**2)


def geodetic2ecef(lat: float, lon: float, alt: float,
                  ell: Ellipsoid=None, deg: bool=True) -> Tuple[float, float, float]:
    """
    Point

    input:
    -----
    lat, lon (degrees)
    alt (altitude, meters)    [0, Infinity)
    ell    reference ellipsoid
    deg    degrees input/output  (False: radians in/out)

    output: ECEF x,y,z (meters)
    """
    if ell is None:
        ell = Ellipsoid()

    if deg:
        lat = radians(lat)
        lon = radians(lon)

    with np.errstate(invalid='ignore'):
        if ((lat < -pi / 2) | (lat > pi / 2)).any():
            raise ValueError('-90 <= lat <= 90')

        if ((lon < -pi) | (lon > 2 * pi)).any():
            raise ValueError('-180 <= lat <= 360')

        if (np.asarray(alt) < 0).any():
            raise ValueError('altitude \in  [0, Infinity)')
    # radius of curvature of the prime vertical section
    N = get_radius_normal(lat, ell)
    # Compute cartesian (geocentric) coordinates given  (curvilinear) geodetic
    # coordinates.
    x = (N + alt) * cos(lat) * cos(lon)
    y = (N + alt) * cos(lat) * sin(lon)
    z = (N * (ell.b / ell.a)**2 + alt) * sin(lat)

    return x, y, z


def ecef2geodetic(x: float, y: float, z: float,
                  ell: Ellipsoid=None, deg: bool=True) -> Tuple[float, float, float]:
    """
    convert ECEF (meters) to geodetic coordinates

    input
    -----
    x,y,z  [meters] target ECEF location                             [0,Infinity)
    ell    reference ellipsoid
    deg    degrees input/output  (False: radians in/out)

    output
    ------
    lat,lon   (degrees/radians)
    alt  (meters)

    Algorithm is based on
    http://www.astro.uni.torun.pl/~kb/Papers/geod/Geod-BG.htm
    This algorithm provides a converging solution to the latitude equation
    in terms of the parametric or reduced latitude form (v)
    This algorithm provides a uniform solution over all latitudes as it does
    not involve division by cos(phi) or sin(phi)
    """
    if ell is None:
        ell = Ellipsoid()

    ea = ell.a
    eb = ell.b
    rad = hypot(x, y)
# Constant required for Latitude equation
    rho = arctan2(eb * z, ea * rad)
# Constant required for latitude equation
    c = (ea**2 - eb**2) / hypot(ea * rad, eb * z)
# Starter for the Newtons Iteration Method
    vnew = arctan2(ea * z, eb * rad)
# Initializing the parametric latitude
    v = 0
    for _ in range(5):
        v = deepcopy(vnew)
# %% Newtons Method for computing iterations
        vnew = v - ((2 * sin(v - rho) - c * sin(2 * v)) /
                    (2 * (cos(v - rho) - c * cos(2 * v))))

        if allclose(v, vnew):
            break
# %% Computing latitude from the root of the latitude equation
    lat = arctan2(ea * tan(vnew), eb)
    # by inspection
    lon = arctan2(y, x)

    alt = (((rad - ea * cos(vnew)) * cos(lat)) +
           ((z - eb * sin(vnew)) * sin(lat)))

    with np.errstate(invalid='ignore'):
        if ((lat < -pi / 2) | (lat > pi / 2)).any():
            raise ValueError('-90 <= lat <= 90')

        if ((lon < -pi) | (lon > 2 * pi)).any():
            raise ValueError('-180 <= lat <= 360')

    if deg:
        return degrees(lat), degrees(lon), alt
    else:
        return lat, lon, alt  # radians


def ecef2enuv(u: float, v: float, w: float,
              lat0: float, lon0: float, deg: bool=True) -> Tuple[float, float, float]:
    """
    for VECTOR i.e. between two points

    input
    -----
    x,y,z  [meters] target ECEF location                             [0,Infinity)

    """
    if deg:
        lat0 = radians(lat0)
        lon0 = radians(lon0)

    t = cos(lon0) * u + sin(lon0) * v
    uEast = -sin(lon0) * u + cos(lon0) * v
    wUp = cos(lat0) * t + sin(lat0) * w
    vNorth = -sin(lat0) * t + cos(lat0) * w

    return uEast, vNorth, wUp


def ecef2enu(x: float, y: float, z: float,
             lat0: float, lon0: float, h0: float,
             ell: Ellipsoid=None, deg: bool=True) -> Tuple[float, float, float]:
    """

    input
    -----
    x,y,z  [meters] target ECEF location                             [0,Infinity)
    Observer: lat0, lon0, h0 (altitude, meters)
    ell    reference ellipsoid
    deg    degrees input/output  (False: radians in/out)

    output:
    -------
    e,n,u   East, North, Up [m]

    """
    x0, y0, z0 = geodetic2ecef(lat0, lon0, h0, ell, deg=deg)

    return uvw2enu(x - x0, y - y0, z - z0, lat0, lon0, deg=deg)


def enu2uvw(east: float, north: float, up: float,
            lat0: float, lon0: float, deg: bool=True) -> Tuple[float, float, float]:
    if deg:
        lat0 = radians(lat0)
        lon0 = radians(lon0)

    t = cos(lat0) * up - sin(lat0) * north
    w = sin(lat0) * up + cos(lat0) * north

    u = cos(lon0) * t - sin(lon0) * east
    v = sin(lon0) * t + cos(lon0) * east

    return u, v, w


def uvw2enu(u: float, v: float, w: float,
            lat0: float, lon0: float, deg: bool=True) -> Tuple[float, float, float]:
    if deg:
        lat0 = radians(lat0)
        lon0 = radians(lon0)

    t = cos(lon0) * u + sin(lon0) * v
    East = -sin(lon0) * u + cos(lon0) * v
    Up = cos(lat0) * t + sin(lat0) * w
    North = -sin(lat0) * t + cos(lat0) * w

    return East, North, Up


def eci2geodetic(eci: np.ndarray, t: datetime,
                 useastropy: bool=True) -> Tuple[float, float, float]:
    """
    convert ECI to geodetic coordinates

    inputs:

    eci/ecef: Nx3 vector of x,y,z triplets in the eci or ecef system [meters]
    t : length N vector of datetime OR greenwich sidereal time angle [radians].


    output
    ------
    lat,lon   (degrees/radians)
    alt  (meters)

    Note: Conversion is idealized: doesn't consider nutations, perterbations,
    etc. like the IAU-76/FK5 or IAU-2000/2006 model-based conversions
    from ECI to ECEF

    eci2geodetic() a.k.a. eci2lla()
    """
    ecef = np.atleast_2d(eci2ecef(eci, t, useastropy))

    return np.asarray(ecef2geodetic(ecef[:, 0], ecef[:, 1], ecef[:, 2])).squeeze()


def enu2ecef(e1: float, n1: float, u1: float,
             lat0: float, lon0: float, h0: float,
             ell: Ellipsoid=None, deg: bool=True) -> Tuple[float, float, float]:
    """
    ENU to ECEF

    inputs:
     e1, n1, u1 (meters)   east, north, up
     observer: lat0, lon0, h0 (degrees/radians,degrees/radians, meters)
    ell    reference ellipsoid
    deg    degrees input/output  (False: radians in/out)


    output
    ------
    x,y,z  [meters] target ECEF location                         [0,Infinity)
    """
    x0, y0, z0 = geodetic2ecef(lat0, lon0, h0, ell, deg=deg)
    dx, dy, dz = enu2uvw(e1, n1, u1, lat0, lon0, deg=deg)

    return x0 + dx, y0 + dy, z0 + dz



from .enu import geodetic2enu, aer2enu, enu2aer
from .ecef import ecef2geodetic, ecef2enuv, ecef2enu, enu2ecef
from typing import Tuple


def aer2ned(az: float, elev: float, slantRange: float, deg: bool=True) -> Tuple[float, float, float]:
    """
    input
    -----
    azimuth, elevation (degrees/radians)                             [0,360),[0,90]
    slant range [meters]                                             [0,Infinity)
    deg    degrees input/output  (False: radians in/out)

    output:
    -------
    n,e,d  North,east, down [m]
    """
    e, n, u = aer2enu(az, elev, slantRange, deg=deg)

    return n, e, -u


def ned2aer(n: float, e: float, d: float, deg: bool=True) -> Tuple[float, float, float]:
    """
    Observer => Point

    input
    -----
    n,e,d  [meters]  North,east, down                                [0,Infinity)
    deg    degrees input/output  (False: radians in/out)

    output: AER
    ------
    azimuth, elevation (degrees/radians)                             [0,360),[0,90]
    slant range [meters]                                             [0,Infinity)
    """
    return enu2aer(e, n, -d, deg=deg)


def ned2geodetic(n: float, e: float, d: float,
                 lat0: float, lon0: float, h0: float,
                 ell=None, deg: bool=True) -> Tuple[float, float, float]:
    """

    input
    -----
    n,e,d   North, east, down (meters)
    Observer: lat0, lon0, h0 (altitude, meters)
    ell    reference ellipsoid
    deg    degrees input/output  (False: radians in/out)


    output:
    -------
    target: lat,lon, h  (degrees/radians,degrees/radians, meters)

    """
    x, y, z = enu2ecef(e, n, -d, lat0, lon0, h0, ell, deg=deg)

    return ecef2geodetic(x, y, z, ell, deg=deg)


def ned2ecef(n: float, e: float, d: float,
             lat0: float, lon0: float, h0: float,
             ell=None, deg: bool=True) -> Tuple[float, float, float]:
    """
    Observer => Point

    input
    -----
    n,e,d  [meters]  North,east, down                                [0,Infinity)
    Observer: lat0, lon0, h0 (altitude, meters)
    ell    reference ellipsoid
    deg    degrees input/output  (False: radians in/out)

    output:
    ------
    ECEF  x,y,z  (meters)
    """
    return enu2ecef(e, n, -d, lat0, lon0, h0, ell, deg=deg)


def ecef2ned(x, y, z, lat0, lon0, h0, ell=None, deg=True) -> Tuple[float, float, float]:
    """
    input
    -----
    x,y,z  [meters] target ECEF location                             [0,Infinity)
    Observer: lat0, lon0, h0 (altitude, meters)
    ell    reference ellipsoid
    deg    degrees input/output  (False: radians in/out)

    output:
    -------
    n,e,d  North,east, down [m]

    """
    e, n, u = ecef2enu(x, y, z, lat0, lon0, h0, ell, deg=deg)

    return n, e, -u


def geodetic2ned(lat, lon, h, lat0, lon0, h0, ell=None, deg=True) -> Tuple[float, float, float]:
    """
    input
    -----
    target: lat,lon  (degrees/radians)
    h (altitude, meters)
    Observer: lat0, lon0 (degrees/radians)
    h0 (altitude, meters)
    ell    reference ellipsoid
    deg    degrees input/output  (False: radians in/out)


    output:
    -------
    n,e,d  North,east, down [m]
    """
    e, n, u = geodetic2enu(lat, lon, h, lat0, lon0, h0, ell, deg=deg)

    return n, e, -u


def ecef2nedv(u, v, w, lat0, lon0, deg=True) -> Tuple[float, float, float]:
    """
    for VECTOR between two points
    """
    e, n, u = ecef2enuv(u, v, w, lat0, lon0, deg=deg)

    return n, e, -u



#!/usr/bin/env python
import pytest
from pytest import approx
import numpy as np
import pymap3d as pm
import pymap3d.datetime2hourangle as pmd
import pymap3d.haversine as pmh

lon = -148
t0 = '2014-04-06T08:00:00'
lla0 = (42, -82, 200)
sra = 2.90658
ha = 45.482789587392013
eci0 = (-3.977913815668146e6, -2.582332196263046e6, 4.250818828152067e6)


def test_datetime2sidereal():
    pytest.importorskip('astropy')
    # http://www.jgiesen.de/astro/astroJS/siderealClock/
    assert pmd.datetime2sidereal(t0, np.radians(lon), False) == approx(sra, rel=1e-5)

    assert pmd.datetime2sidereal([t0], np.radians(lon), False) == approx([sra], rel=1e-5)


def test_datetime2sidereal_vallado():
    assert pmd.datetime2sidereal(t0, np.radians(lon), True) == approx(sra, rel=1e-5)

    assert pmd.datetime2sidereal([t0], np.radians(lon), True) == approx([sra], rel=1e-5)


def test_anglesep():
    pytest.importorskip('astropy')

    assert pmh.anglesep(35, 23, 84, 20) == approx(ha)


def test_anglesep_meeus():
    # %% compare with astropy
    assert pmh.anglesep_meeus(35, 23, 84, 20) == approx(ha)


def test_eci_astropy():
    pytest.importorskip('astropy')

    t = '2013-01-15T12:00:05'
    lla = pm.eci2geodetic(eci0, t)
    assert lla == approx(lla0, rel=0.2)

    eci1 = pm.eci2ecef(eci0, t)
    assert eci1 == approx([649012.04640917, -4697980.55129606, 4250818.82815207])

    assert pm.ecef2eci(eci1, t) == approx(eci0)

    aer1 = pm.eci2aer(eci0, 42, -100, 0, t)
    assert aer1 == approx([83.73050, -6.614478, 1.473510e6])

    assert pm.aer2eci(*aer1, 42, -100, 0, t) == approx(eci0)

    with pytest.raises(ValueError):
        pm.aer2eci(aer1[0], aer1[1], -1, 42, -100, 0, t)


def test_eci_times_astropy():
    pytest.importorskip('astropy')

    with pytest.raises(AssertionError):
        pm.eci2ecef(eci0, [t0, t0])

    with pytest.raises(AssertionError):
        pm.ecef2eci(eci0, [t0, t0])

    eci0s = np.stack((eci0, eci0))
    assert pm.ecef2eci(pm.eci2ecef(eci0s, [t0] * 2), [t0] * 2) == approx(eci0s)


def test_eci_vallado():
    t = '2013-01-15T12:00:05'
    lla = pm.eci2geodetic(eci0, t, useastropy=False)
    assert lla == approx(lla0, rel=0.2)

    eci1 = pm.eci2ecef(eci0, t, useastropy=False)
    assert eci1 == approx([649012.04640917, -4697980.55129606, 4250818.82815207], rel=0.001)

    assert pm.ecef2eci(eci1, t, useastropy=False) == approx(eci0, rel=0.001)

    aer1 = pm.eci2aer(eci0, 42, -100, 0, t, useastropy=False)
    assert aer1 == approx([83.73050, -6.614478, 1.473510e6], rel=0.001)

    assert pm.aer2eci(*aer1, 42, -100, 0, t, useastropy=False) == approx(eci0, rel=0.001)

    with pytest.raises(ValueError):
        pm.aer2eci(aer1[0], aer1[1], -1, 42, -100, 0, t, useastropy=False)


def test_eci_times_vallado():
    with pytest.raises(AssertionError):
        pm.eci2ecef(eci0, [t0, t0], useastropy=False)

    with pytest.raises(AssertionError):
        pm.ecef2eci(eci0, [t0, t0], useastropy=False)

    eci0s = np.stack((eci0, eci0))
    assert pm.ecef2eci(pm.eci2ecef(eci0s, [t0] * 2, useastropy=False),
                       [t0] * 2, useastropy=False) == approx(eci0s, rel=0.001)


if __name__ == '__main__':
    pytest.main(['-xrsv', __file__])



# Copyright (c) 2014-2018 Michael Hirsch, Ph.D.
"""
 Ported by Michael Hirsch to Python.
"""
from typing import Tuple
import logging
from numpy import arctan, sqrt, tan, sign, sin, cos, arctan2, arcsin, nan, pi
import numpy as np
from .ecef import Ellipsoid


def vdist(Lat1: float, Lon1: float, Lat2: float, Lon2: float, ell: Ellipsoid=None) -> Tuple[float, float, float]:
    """
    Using the WGS-84 Earth ellipsoid, compute the distance between two points
    within a few millimeters of accuracy, compute forward azimuth,
    and compute backward azimuth, all using a vectorized version of

    Vincenty's algorithm::

        s = vdist(lat1,lon1,lat2,lon2)
        s,a12 = vdist(lat1,lon1,lat2,lon2)
        s,a12,a21 = vdist(lat1,lon1,lat2,lon2)


    s
         distance in meters (inputs may be scalars, vectors, or matrices)

    a12
         azimuth in degrees from first point to second point (forward)

    a21
         azimuth in degrees from second point to first point (backward)

    (Azimuths are in degrees clockwise from north.)

    lat1
        GEODETIC latitude of first point (degrees)

    lon1
        longitude of first point (degrees)

    lat2, lon2
        second point(s) (degrees)

    Original algorithm source:
    T. Vincenty, "Direct and Inverse Solutions of Geodesics on the Ellipsoid
    with Application of Nested Equations", Survey Review, vol. 23, no. 176,
    April 1975, pp 88-93.
    Available at: http://www.ngs.noaa.gov/PUBS_LIB/inverse.pdf

    Notes:
     1. lat1,lon1,lat2,lon2 can be any (identical) size/shape. Outputs  will have the same size and shape.
     2. Error correcting code, convergence failure traps, antipodal corrections, polar error corrections, WGS84 ellipsoid parameters, testing, and comments: Michael Kleder, 2004.
     3. Azimuth implementation (including quadrant abiguity resolution) and code vectorization, Michael Kleder, Sep 2005.
     4. Vectorization is convergence sensitive; that is, quantities which have already converged to within tolerance are not recomputed during subsequent iterations (while other quantities are still converging).
     5. Vincenty describes his distance algorithm as precise to within 0.01 millimeters, subject to the ellipsoidal model.
     6. For distance calculations, essentially antipodal points are treated as exactly antipodal, potentially reducing accuracy slightly.
     7. Distance failures for points exactly at the poles are eliminated by moving the points by 0.6 millimeters.
     8. The Vincenty distance algorithm was transcribed verbatim by Peter Cederholm, August 12, 2003. It was modified and translated to English by Michael Kleder. Mr. Cederholm's website is http://www.plan.aau.dk/~pce/
     9. Distances agree with the Mapping Toolbox, version 2.2 (R14SP3) with a max relative difference of about 5e-9, except when the two points are nearly antipodal, and except when one point is near the equator and the two longitudes are nearly 180 degrees apart. This function (vdist) is more accurate in such cases.
        For example, note this difference (as of this writing)::

           vdist(0.2,305,15,125)

        > 18322827.0131551

        ::

            distance(0.2,305,15,125,[6378137 0.08181919])

        > 0
     10. Azimuths FROM the north pole (either forward starting at the north pole or backward when ending at the north pole) are set to 180 degrees by convention.
         Azimuths FROM the south pole are set to 0 degrees by convention.
     11. Azimuths agree with the Mapping Toolbox, version 2.2 (R14SP3) to within about a hundred-thousandth of a degree, except when traversing to or from a pole, where the convention for this function is described in (10), and except in the cases noted above in (9).
     12. No warranties; use at your own risk.

    """
    if ell is None:
        ell = Ellipsoid()
# %% prepare inputs
    lat1 = np.atleast_1d(Lat1)
    lat2 = np.atleast_1d(Lat2)
    lon1 = np.atleast_1d(Lon1)
    lon2 = np.atleast_1d(Lon2)

    assert lat1.shape == lon1.shape and lat2.shape == lon2.shape

    if lat1.shape != lat2.shape:
        if lat1.size == 1:
            lat1 = np.broadcast_to(lat1, lat2.shape)
            lon1 = np.broadcast_to(lon1, lon2.shape)

        if lat2.size == 1:
            lat2 = np.broadcast_to(lat2, lat1.shape)
            lon2 = np.broadcast_to(lon2, lon1.shape)
# %% Input check:
    if ((abs(lat1) > 90) | (abs(lat2) > 90)).any():
        raise ValueError('Input latitudes must be in [-90, 90] degrees.')
# %% Supply WGS84 earth ellipsoid axis lengths in meters:
    a = ell.a
    b = ell.b
# %% preserve true input latitudes:
    lat1tr = lat1.copy()
    lat2tr = lat2.copy()
# %% convert inputs in degrees to np.radians:
    lat1 = np.radians(lat1)
    lon1 = np.radians(lon1)
    lat2 = np.radians(lat2)
    lon2 = np.radians(lon2)
# %% correct for errors at exact poles by adjusting 0.6 millimeters:
    kidx = abs(pi / 2 - abs(lat1)) < 1e-10
    if kidx.any():
        lat1[kidx] = sign(lat1[kidx]) * (pi / 2 - (1e-10))

    kidx = abs(pi / 2 - abs(lat2)) < 1e-10
    if kidx.any():
        lat2[kidx] = sign(lat2[kidx]) * (pi / 2 - (1e-10))

    f = (a - b) / a
    U1 = arctan((1 - f) * tan(lat1))
    U2 = arctan((1 - f) * tan(lat2))
    lon1 = lon1 % (2 * pi)
    lon2 = lon2 % (2 * pi)
    L = abs(lon2 - lon1)
    kidx = L > pi
    if kidx.any():
        L[kidx] = 2 * pi - L[kidx]

    lamb = L.copy()  # NOTE: program will fail without copy!
    lambdaold = np.zeros(lat1.shape)
    itercount = 0
    notdone = np.ones(lat1.shape, dtype=bool)
    alpha = np.zeros(lat1.shape)
    sigma = np.zeros(lat1.shape)
    cos2sigmam = np.zeros(lat1.shape)
    C = np.zeros(lat1.shape)
    warninggiven = False
    sinsigma = np.empty(notdone.shape)
    cossigma = np.empty(notdone.shape)
    while notdone.any():  # force at least one execution
        itercount += 1
        if itercount > 50:
            if not warninggiven:
                logging.warning('Essentially antipodal points--precision may be reduced slightly.')

            lamb[notdone] = pi
            break

        lambdaold[notdone] = lamb[notdone]

        sinsigma[notdone] = sqrt(
            (cos(U2[notdone]) * sin(lamb[notdone]))**2 +
            (cos(U1[notdone]) * sin(U2[notdone]) - sin(U1[notdone]) *
             cos(U2[notdone]) * cos(lamb[notdone]))**2)

        cossigma[notdone] = (sin(U1[notdone]) * sin(U2[notdone]) +
                             cos(U1[notdone]) * cos(U2[notdone]) *
                             cos(lamb[notdone]))
        # eliminate rare imaginary portions at limit of numerical precision:
        sinsigma[notdone] = sinsigma[notdone].real
        cossigma[notdone] = cossigma[notdone].real

        sigma[notdone] = arctan2(sinsigma[notdone], cossigma[notdone])

        alpha[notdone] = (arcsin(cos(U1[notdone]) * cos(U2[notdone]) *
                                 sin(lamb[notdone]) / sin(sigma[notdone])))

        cos2sigmam[notdone] = (cos(sigma[notdone]) - 2 * sin(U1[notdone]) *
                               sin(U2[notdone]) / cos(alpha[notdone])**2)

        C[notdone] = (f / 16 * cos(alpha[notdone])**2 *
                      (4 + f * (4 - 3 * cos(alpha[notdone])**2)))

        lamb[notdone] = (L[notdone] + (1 - C[notdone]) * f * sin(alpha[notdone]) *
                         (sigma[notdone] + C[notdone] * sin(sigma[notdone]) *
                          (cos2sigmam[notdone] + C[notdone] * cos(sigma[notdone]) *
                           (-1 + 2. * cos2sigmam[notdone]**2))))
        # print(f'then, lambda(21752) = {lamb[21752],20})
        # correct for convergence failure for essentially antipodal points
        if (lamb[notdone] > pi).any():
            logging.warning('Essentially antipodal points encountered. Precision may be reduced slightly.')
            warninggiven = True
            lambdaold[lamb > pi] = pi
            lamb[lamb > pi] = pi

        notdone = abs(lamb - lambdaold) > 1e-12

    u2 = cos(alpha)**2 * (a**2 - b**2) / b**2
    A = 1 + u2 / 16384 * (4096 + u2 * (-768 + u2 * (320 - 175 * u2)))
    B = u2 / 1024 * (256 + u2 * (-128 + u2 * (74 - 47 * u2)))
    deltasigma = (B * sin(sigma) *
                  (cos2sigmam + B / 4 * (cos(sigma) * (-1 + 2 * cos2sigmam**2) -
                                         B / 6 * cos2sigmam * (-3 + 4 * sin(sigma)**2) * (-3 + 4 * cos2sigmam**2))))

    dist_m = (b * A * (sigma - deltasigma))

# %% From point #1 to point #2
    # correct sign of lambda for azimuth calcs:
    lamb = abs(lamb)
    kidx = sign(sin(lon2 - lon1)) * sign(sin(lamb)) < 0
    lamb[kidx] = -lamb[kidx]
    numer = cos(U2) * sin(lamb)
    denom = cos(U1) * sin(U2) - sin(U1) * cos(U2) * cos(lamb)
    a12 = arctan2(numer, denom)
    kidx = a12 < 0
    a12[kidx] = a12[kidx] + 2 * pi
    # %% from poles
    a12[lat1tr <= -90] = 0
    a12[lat1tr >= 90] = pi
    az = np.degrees(a12)

# %% From point #2 to point #1
    # correct sign of lambda for azimuth calcs:
    lamb = abs(lamb)
    kidx = sign(sin(lon1 - lon2)) * sign(sin(lamb)) < 0
    lamb[kidx] = -lamb[kidx]
    numer = cos(U1) * sin(lamb)
    denom = sin(U1) * cos(U2) - cos(U1) * sin(U2) * cos(lamb)
    a21 = arctan2(numer, denom)
    kidx = a21 < 0
    a21[kidx] = a21[kidx] + 2 * pi
    # %% backwards from poles:
    a21[lat2tr >= 90] = pi
    a21[lat2tr <= -90] = 0.
    backaz = np.degrees(a21)

    return dist_m.squeeze()[()], az.squeeze()[()], backaz.squeeze()[()]


def vreckon(Lat1: float, Lon1: float, Rng: float, Azim: float,
            ell: Ellipsoid=None) -> Tuple[float, float, float]:
    """
    This is the Vincenty "forward" solution.

    Computes points at a specified azimuth and range in an ellipsoidal earth.
    Using the WGS-84 Earth ellipsoid, travel a given distance along a given azimuth starting at a given initial point,
    and return the endpoint within a few millimeters of accuracy, using Vincenty's algorithm.

    USAGE::

        lat2,lon2 = vreckon(lat1, lon1, rng, azim)

    Transmits ellipsoid definition (either as [a,b] or [a,f]) as fifth argument ELLIPSOID


    VARIABLES

    lat1
        inital latitude (degrees)

    lon1
        initial longitude (degrees)

    rng
        distance (meters). Scalar or a vector. Latter case computes a series of circles (or arc circles, see azim) centered on X,Y (which are scalars)

    azim
        intial azimuth (degrees). "azim" is a scalar or vector

    ellipsoid
        two-element ellipsoid vector. Either [a b] or [a f] If omitted, defaults to WGS-84

    lat2, lon2
        second point (degrees)

    a21
        reverse azimuth (degrees), at final point facing back toward the intial point

    :Original algorithm: T. Vincenty, "Direct and Inverse Solutions of Geodesics on the Ellipsoid with Application of Nested Equations", Survey Review, vol. 23, no. 176, April 1975, pp 88-93. http://www.ngs.noaa.gov/PUBS_LIB/inverse.pdf

    Notes:

      1. The Vincenty reckoning algorithm was transcribed verbatim into JavaScript by Chris Veness.
         It was modified and translated to Matlab by Michael Kleder.
         Mr. Veness's website is: http://www.movable-type.co.uk/scripts/latlong-vincenty-direct.html
      2. Error correcting code, polar error corrections, WGS84 ellipsoid parameters, testing, and comments by Michael Kleder.
      3. By convention, when starting at a pole, the longitude of the initial point (otherwise meaningless) determines the longitude line along which to traverse, and hence the longitude of the final point.
      4. The convention noted in (3) above creates a discrepancy with VDIST when the the intial or final point is at a pole. In the VDIST
         function, when traversing from a pole, the azimuth is  0 when
         heading away from the south pole and 180 when heading away from the north pole. In contrast, this VRECKON function uses the azimuth as noted in (3) above when traversing away form a pole.
      5. In testing, where the traversal subtends no more than 178 degrees, this function correctly inverts the VDIST function to within 0.2 millimeters of distance, 5e-10 degrees of forward azimuth,
         and 5e-10 degrees of reverse azimuth. Precision reduces as test
         points approach antipodal because the precision of VDIST is reduced
         for nearly antipodal points. (A warning is given by VDIST.)
      6. Tested but no warranty. Use at your own risk.
      7. Ver 1.0, Michael Kleder, November 2007. Ver 2.0, Joaquim Luis, September 2008

    Added ellipsoid and vectorized whenever possible. Also, lon2 is always converted to the [-180 180] interval.
    Joaquim Luis

    """

    lat1 = np.atleast_1d(Lat1)
    lon1 = np.atleast_1d(Lon1)
    rng = np.atleast_1d(Rng)
    azim = np.atleast_1d(Azim)

    if rng.ndim != 1 or azim.ndim != 1:
        raise ValueError('Range and azimuth must be scalar or vector')

    if abs(lat1) > 90:
        raise ValueError('VRECKON: Input lat. must be between -90 and 90 deg., inclusive.')

    if lat1.size > 1 and rng.size > 1:
        raise ValueError('VRECKON: Variable ranges are only allowed for a single point.')

    if ell is not None:
        a = ell.a
        b = ell.b
        f = ell.f
    else:   # Supply WGS84 earth ellipsoid axis lengths in meters:
        a = 6378137                 # semimajor axis
        b = 6356752.31424518   # WGS84 earth flattening coefficient definition
        f = (a - b) / a

    lat1 = np.radians(lat1)            # intial latitude in radians
    lon1 = np.radians(lon1)            # intial longitude in radians

    # correct for errors at exact poles by adjusting 0.6 millimeters:
    kidx = abs(pi / 2 - abs(lat1)) < 1e-10
    lat1[kidx] = sign(lat1[kidx]) * (pi / 2 - (1e-10))

    if rng.size != azim.size and rng.size == 1:
        rng = np.broadcast_to(rng, azim.size)

    alpha1 = np.radians(azim)  # inital azimuth in radians
    sinAlpha1 = sin(alpha1)
    cosAlpha1 = cos(alpha1)

    tanU1 = (1 - f) * tan(lat1)
    cosU1 = 1 / sqrt(1 + tanU1 ** 2)
    sinU1 = tanU1 * cosU1
    sigma1 = arctan2(tanU1, cosAlpha1)
    sinAlpha = cosU1 * sinAlpha1
    cosSqAlpha = 1 - sinAlpha * sinAlpha
    uSq = cosSqAlpha * (a**2 - b**2) / b**2
    A = 1 + uSq / 16384 * (4096 + uSq * (-768 + uSq * (320 - 175 * uSq)))
    B = uSq / 1024 * (256 + uSq * (-128 + uSq * (74 - 47 * uSq)))

    sigma = rng / (b * A)
    sigmaP = 2 * pi

    if sigma.size == 1:
        sinSigma = nan
        cosSigma = nan
        cos2SigmaM = nan
        while abs(sigma - sigmaP) > 1e-12:
            cos2SigmaM = cos(2 * sigma1 + sigma)
            sinSigma = sin(sigma)
            cosSigma = cos(sigma)
            deltaSigma = (B * sinSigma *
                          (cos2SigmaM + B / 4 *
                           (cosSigma * (-1 + 2 * cos2SigmaM * cos2SigmaM) -
                            B / 6 * cos2SigmaM *
                            (-3 + 4 * sinSigma * sinSigma) *
                            (-3 + 4 * cos2SigmaM * cos2SigmaM))))
            sigmaP = sigma
            sigma = rng / (b * A) + deltaSigma
    else:
        # This part is not vectorized
        cos2SigmaM = np.empty(sigma.size)
        sinSigma = np.empty(sigma.size)
        cosSigma = np.empty(sigma.size)

        for k in range(sigma.size):
            while (abs(sigma[k] - sigmaP) > 1e-12).any():
                cos2SigmaM[k] = cos(2 * sigma1[k] + sigma[k])
                sinSigma[k] = sin(sigma[k])
                cosSigma[k] = cos(sigma[k])
                tmp = 2 * cos2SigmaM[k] * cos2SigmaM[k]
                deltaSigma = (B[k] * sinSigma[k] *
                              (cos2SigmaM[k] + B[k] / 4 *
                               (cosSigma[k] * (-1 + tmp) -
                                B[k] / 6 * cos2SigmaM[k] *
                                (-3 + 4 * sinSigma[k] * sinSigma[k]) *
                                  (-3 + 2 * tmp))))
                sigmaP = sigma[k]
                sigma[k] = rng[k] / (b * A[k]) + deltaSigma

    tmp = sinU1 * sinSigma - cosU1 * cosSigma * cosAlpha1
    lat2 = arctan2(sinU1 * cosSigma + cosU1 * sinSigma * cosAlpha1,
                   (1 - f) * sqrt(sinAlpha * sinAlpha + tmp**2))

    lamb = arctan2(sinSigma * sinAlpha1,
                   cosU1 * cosSigma - sinU1 * sinSigma * cosAlpha1)

    C = f / 16 * cosSqAlpha * (4 + f * (4 - 3 * cosSqAlpha))
    L = lamb - (f * (1 - C) *
                sinAlpha * (sigma + C * sinSigma *
                            (cos2SigmaM + C * cosSigma *
                             (-1 + 2 * cos2SigmaM * cos2SigmaM))))

    lon2 = np.degrees(lon1 + L)

    # Truncates angles into the [-pi pi] range
    # if (lon2 > pi).any():
    #    lon2 = pi*((absolute(lon2)/pi) -
    #       2*ceil(((absolute(lon2)/pi)-1)/2)) * sign(lon2)

    # lon2 = mod(lon2,360); % follow [0,360] convention
    lon2 = (lon2 + 180) % 360 - 180  # no parenthesis on RHS

    a21 = arctan2(sinAlpha, -tmp)
    a21 = 180. + np.degrees(a21)  # note direction reversal

    return np.degrees(lat2).squeeze()[()], lon2.squeeze()[()], a21.squeeze()[()] % 360.


def track2(lat1: float, lon1: float, lat2: float, lon2: float,
           ell: Ellipsoid=None, npts: int=100, deg: bool=True):
    """
     computes great circle tracks starting at the point lat1, lon1 and ending at lat2, lon2

     input
     -----
     lat1
     GEODETIC latitude of first point (degrees/radians)

     lon1
     longitude of first point (degrees/radians)

     lat2, lon2
     second point (degrees/radians)

     ell    reference ellipsoid
     npts   number of points (default is 100)
     deg    degrees input/output  (False: radians in/out)

     output
     ------
     lats, lons  latitudes and longitudes of points along great circle track

     Based on code posted to the GMT mailing list in Dec 1999 by Jim Levens and by Jeff Whitaker <jeffrey.s.whitaker@noaa.gov>
     """

    if ell is None:
        ell = Ellipsoid()

    if npts <= 1:
        raise ValueError('npts must be greater than 1')

    if npts == 2:
        return [lat1, lat2], [lon1, lon2]

    if deg is True:
        rlat1, rlon1, rlat2, rlon2 = np.radians([lat1, lon1, lat2, lon2])
    else:
        rlat1, rlon1, rlat2, rlon2 = lat1, lon1, lat2, lon2

    gcarclen = 2. * np.arcsin(np.sqrt((np.sin((rlat1 - rlat2) / 2))**2 +
                                      np.cos(rlat1) * np.cos(rlat2) * (np.sin((rlon1 - rlon2) / 2))**2))
    # check to see if points are antipodal (if so, route is undefined).
    if np.allclose(gcarclen, pi):
        raise ValueError('cannot compute intermediate points on a great circle whose endpoints are antipodal')

    distance, azimuth, _ = vdist(lat1, lon1, lat2, lon2)
    incdist = distance / (npts - 1)

    latpt = lat1
    lonpt = lon1
    lons = [lonpt]
    lats = [latpt]
    for n in range(npts - 2):
        latptnew, lonptnew, _ = vreckon(latpt, lonpt, incdist, azimuth)
        _, azimuth, _ = vdist(latptnew, lonptnew, lat2, lon2, ell=ell)
        lats.append(latptnew)
        lons.append(lonptnew)
        latpt = latptnew
        lonpt = lonptnew
    lons.append(lon2)
    lats.append(lat2)

    if not deg:
        lats = np.radians(lats)
        lons = np.radians(lons)

    return lats, lons



# Copyright (c) 2014-2018 Michael Hirsch, Ph.D.
from datetime import datetime
from dateutil.parser import parse
import numpy as np


def str2dt(time: datetime) -> np.ndarray:
    """
    Converts times in string or list of strings to datetime(s)

    output: datetime
    """
    if isinstance(time, datetime):
        return time
    elif isinstance(time, str):
        return parse(time)
    else:  # some sort of iterable
        try:
            return [parse(t) for t in time]
        except TypeError:
            return time.values.astype('datetime64[us]').astype(datetime)



#!/usr/bin/env python
# Copyright (c) 2014-2018 Michael Hirsch, Ph.D.

"""
runs tests
"""
import pytest
from pytest import approx
import numpy as np
import pymap3d as pm

pi = np.pi
nan = np.nan

lla0 = (42, -82, 200)
rlla0 = (np.radians(lla0[0]), np.radians(lla0[1]), lla0[2])

aer0 = (33, 70, 1000)
raer0 = (np.radians(aer0[0]), np.radians(aer0[1]), aer0[2])

# %% outcomes from matlab
xyz0 = (660.6753e3, -4700.9487e3, 4245.738e3)  # geodetic2ecef

lla1 = (42.002582, -81.997752, 1.1397018e3)  # aer2geodetic
rlla1 = (np.radians(lla1[0]), np.radians(lla1[1]), lla1[2])

axyz0 = 660930.2, -4701424, 4246579.6  # aer2ecef

enu0 = (186.277521, 286.842228, 939.692621)  # aer2enu
ned0 = (enu0[1], enu0[0], -enu0[2])

# vector
vx, vy, vz = (5, 3, 2)
ve, vn, vu = (5.368859646588048, 3.008520763668120, -0.352347711524077)


def test_ellipsoid():

    assert pm.ecef2geodetic(*xyz0, ell=pm.Ellipsoid('wgs84')) == approx([42., -82., 200.24339])
    assert pm.ecef2geodetic(*xyz0, ell=pm.Ellipsoid('grs80')) == approx([42., -82., 200.24344])
    assert pm.ecef2geodetic(*xyz0, ell=pm.Ellipsoid('clrk66')) == approx([42.00213, -82., 237.17182])
    assert pm.ecef2geodetic(*xyz0, ell=pm.Ellipsoid('mars')) == approx([41.99476, -82., 2.981169e6])
    assert pm.ecef2geodetic(*xyz0, ell=pm.Ellipsoid('venus')) == approx([41.808706, -82., 3.178069e5])
    assert pm.ecef2geodetic(*xyz0, ell=pm.Ellipsoid('moon')) == approx([41.808706, -82., 4.630807e6])


def test_losint():

    pytest.importorskip('pytest', minversion='3.5')

    az = [0., 10., 125.]

    lat, lon, sr = pm.lookAtSpheroid(*lla0, az, tilt=0.)
    assert (lat[0] == lat).all() and (lon[0] == lon).all() and (sr[0] == sr).all()

    assert (lat[0], lon[0], sr[0]) == approx(lla0)

    with pytest.raises(ValueError):
        pm.lookAtSpheroid(lla0[0], lla0[1], -1, az, 0)

# %%
    tilt = [30., 45., 90.]
    lat, lon, sr = pm.lookAtSpheroid(*lla0, az, tilt)

    truth = np.array([[42.00103959, lla0[1], 230.9413173],
                      [42.00177328, -81.9995808, 282.84715651],
                      [nan, nan, nan]])

    assert np.column_stack((lat, lon, sr)) == approx(truth, nan_ok=True)


def test_geodetic():
    xyz = pm.geodetic2ecef(*lla0)

    assert xyz == approx(xyz0)
    assert pm.geodetic2ecef(*rlla0, deg=False) == approx(xyz)

    with pytest.raises(ValueError):
        pm.geodetic2ecef(lla0[0], lla0[1], -1)

    with pytest.raises(ValueError):
        pm.geodetic2ecef(-100, lla0[1], lla0[2])

    with pytest.raises(ValueError):
        pm.geodetic2ecef(lla0[0], -200, lla0[2])

    assert pm.ecef2geodetic(*xyz) == approx(lla0)
    assert pm.ecef2geodetic(*xyz, deg=False) == approx(rlla0)

    lla2 = pm.aer2geodetic(*aer0, *lla0)
    rlla2 = pm.aer2geodetic(*raer0, *rlla0, deg=False)

    with pytest.raises(ValueError):
        pm.aer2geodetic(aer0[0], aer0[1], -1, *lla0)

    assert lla2 == approx(lla1)
    assert rlla2 == approx(rlla1)

    assert pm.geodetic2aer(*lla2, *lla0) == approx(aer0)
    assert pm.geodetic2aer(*rlla2, *rlla0, deg=False) == approx(raer0)

    anan = np.empty((10, 10))
    anan.fill(np.nan)
    assert np.isnan(pm.geodetic2aer(anan, anan, anan, *lla0)).all()
    assert np.isnan(pm.aer2geodetic(anan, anan, anan, *lla0)).all()


def test_aer_ecef():
    xyz = pm.aer2ecef(*aer0, *lla0)

    assert xyz == approx(axyz0)
    assert pm.aer2ecef(*raer0, *rlla0, deg=False) == approx(axyz0)

    with pytest.raises(ValueError):
        pm.aer2ecef(aer0[0], aer0[1], -1, *lla0)

    assert pm.ecef2aer(*xyz, *lla0) == approx(aer0)
    assert pm.ecef2aer(*xyz, *rlla0, deg=False) == approx(raer0)


def test_aer_enu():
    xyz = pm.aer2ecef(*aer0, *lla0)

    enu = pm.aer2enu(*aer0)

    assert enu == approx(enu0)
    assert pm.aer2enu(*raer0, deg=False) == approx(enu0)

    with pytest.raises(ValueError):
        pm.aer2enu(aer0[0], aer0[1], -1)

    assert pm.enu2ecef(*enu, *lla0) == approx(xyz)
    assert pm.enu2ecef(*enu, *rlla0, deg=False) == approx(xyz)

    assert pm.ecef2enu(*xyz, *lla0) == approx(enu)
    assert pm.ecef2enu(*xyz, *rlla0, deg=False) == approx(enu)


def test_ned():
    xyz = pm.aer2ecef(*aer0, *lla0)
    enu = pm.aer2enu(*aer0)
    ned = (enu[1], enu[0], -enu[2])
    lla = pm.aer2geodetic(*aer0, *lla0)

    assert pm.aer2ned(*aer0) == approx(ned0)

    with pytest.raises(ValueError):
        pm.aer2ned(aer0[0], aer0[1], -1)

    assert pm.enu2aer(*enu) == approx(aer0)
    assert pm.enu2aer(*enu, deg=False) == approx(raer0)

    assert pm.ned2aer(*ned) == approx(aer0)

    assert pm.ecef2ned(*xyz, *lla0) == approx(ned)

    assert pm.ned2ecef(*ned, *lla0) == approx(xyz)
# %%
    assert pm.ecef2enuv(vx, vy, vz, *lla0[:2]) == approx((ve, vn, vu))

    assert pm.ecef2nedv(vx, vy, vz, *lla0[:2]) == approx((vn, ve, -vu))

# %%
    enu3 = pm.geodetic2enu(*lla, *lla0)
    ned3 = (enu3[1], enu3[0], -enu3[2])

    assert pm.geodetic2ned(*lla, *lla0) == approx(ned3)

    assert pm.enu2geodetic(*enu3, *lla0) == approx(lla)

    assert pm.ned2geodetic(*ned3, *lla0) == approx(lla)


if __name__ == '__main__':
    pytest.main(['-xrsv', __file__])



#!/usr/bin/env python
"""
Example Kitt Peak

./radec2azel.py 257.96295344 15.437854 31.9583 -111.5967 2014-12-25T22:00:00MST
"""
from pymap3d import radec2azel
from argparse import ArgumentParser


def main():
    p = ArgumentParser(description="RightAscension,Declination =>"
                                   "Azimuth,Elevation")
    p.add_argument('ra', help='right ascension [degrees]', type=float)
    p.add_argument('dec', help='declination [degrees]', type=float)
    p.add_argument('lat', help='WGS84 latitude of observer [degrees]',
                   type=float)
    p.add_argument('lon', help='WGS84 latitude of observer [degrees]',
                   type=float)
    p.add_argument('time', help='UTC time of observation YYYY-mm-ddTHH:MM:SSZ')
    P = p.parse_args()

    az_deg, el_deg = radec2azel(P.ra, P.dec, P.lat, P.lon, P.time)
    print('azimuth: [deg]', az_deg)
    print('elevation [deg]:', el_deg)


if __name__ == '__main__':
    main()



from datetime import datetime
import numpy as np
from .datetime2hourangle import datetime2sidereal
try:
    from astropy.time import Time
except ImportError as e:
    Time = None


def eci2ecef(eci: np.ndarray,
             time: datetime,
             useastropy: bool=True) -> np.ndarray:
    """
     Observer => Point

    input
    -----
    eci [meters] Nx3 target ECI location (x,y,z)                    [0,Infinity)
    t  time (datetime.datetime)   time of obsevation (UTC)

    output
    ------
    x,y,z  [meters] target ECEF location                             [0,Infinity)
    """
    useastropy = useastropy and Time

    if useastropy:
        gst = Time(time).sidereal_time('apparent', 'greenwich').radian
    else:
        gst = datetime2sidereal(time, 0.)

    gst = np.atleast_1d(gst)
    assert gst.ndim == 1 and isinstance(gst[0], float)  # must be in radians!

    eci = np.atleast_2d(eci)
    assert eci.shape[0] == gst.size, 'length of time does not match number of ECI positions'

    N, trip = eci.shape
    if eci.ndim > 2 or trip != 3:
        raise ValueError('eci triplets must be shape (N,3)')

    ecef = np.empty_like(eci)

    for i in range(N):
        ecef[i, :] = _rottrip(gst[i]) @ eci[i, :]

    return ecef.squeeze()


def ecef2eci(ecef: np.ndarray,
             time: datetime,
             useastropy: bool=True) -> np.ndarray:
    """
    Point => Point

    input
    -----
    ecef:  Nx3  x,y,z  (meters)
    time:  datetime.datetime


    output
    ------
    eci  x,y,z (meters)
    """
    useastropy = useastropy and Time

    if useastropy:
        gst = Time(time).sidereal_time('apparent', 'greenwich').radian
    else:
        gst = datetime2sidereal(time, 0.)

    gst = np.atleast_1d(gst)
    assert gst.ndim == 1 and isinstance(gst[0], float)  # must be in radians!

    ecef = np.atleast_2d(ecef)
    assert ecef.shape[0] == gst.size, 'length of time does not match number of ECEF positions'

    N, trip = ecef.shape
    if ecef.ndim > 2 or trip != 3:
        raise ValueError('ecef triplets must be shape (N,3)')

    eci = np.empty_like(ecef)
    for i in range(N):
        eci[i, :] = _rottrip(gst[i]).T @ ecef[i, :]  # this one is transposed

    return eci.squeeze()


def _rottrip(ang: np.ndarray) -> np.ndarray:
    ang = ang.squeeze()
    if ang.size > 1:
        raise ValueError('only one angle allowed at a time')
    """ported from:
    https://github.com/dinkelk/astrodynamics/blob/master/rot3.m
    """
    return np.array([[np.cos(ang), np.sin(ang), 0],
                     [-np.sin(ang), np.cos(ang), 0],
                     [0, 0, 1]])



# Copyright (c) 2014-2018 Michael Hirsch, Ph.D.
from math import pi
from datetime import datetime
import numpy as np
from .timeconv import str2dt
try:
    from astropy.time import Time
    import astropy.units as u
    from astropy.coordinates import Longitude
except ImportError:
    Time = None

"""
The "usevallado" datetime to julian runs 4 times faster than astropy.
However, AstroPy is more accurate.
"""


def datetime2sidereal(time: datetime,
                      lon_radians: float,
                      usevallado: bool=True) -> float:
    """
    Convert ``datetime`` to sidereal time

    :algorithm: D. Vallado Fundamentals of Astrodynamics and Applications


    time
        Python datetime

    lon
        longitude in RADIANS
    """
    usevallado = usevallado or Time is None
    if usevallado:
        jd = datetime2julian(str2dt(time))
# %% Greenwich Sidereal time RADIANS
        gst = np.atleast_1d(julian2sidereal(jd))
# %% Algorithm 15 p. 188 rotate GST to LOCAL SIDEREAL TIME
        tsr = gst + lon_radians
    else:
        tsr = Time(time).sidereal_time(kind='apparent',
                                       longitude=Longitude(lon_radians, unit=u.radian)).radian

    return tsr


def datetime2julian(time: datetime) -> float:
    """
    Python datetime to Julian time

    from D.Vallado Fundamentals of Astrodynamics and Applications p.187
     and J. Meeus Astronomical Algorithms 1991 Eqn. 7.1 pg. 61
    """

    times = np.atleast_1d(time)
    assert times.ndim == 1

    jd = np.empty(times.size)
    for i, t in enumerate(times):
        if t.month < 3:
            year = t.year - 1
            month = t.month + 12
        else:
            year = t.year
            month = t.month

        A = int(year / 100.0)
        B = 2 - A + int(A / 4.)
        C = ((t.second / 60. + t.minute) / 60. + t.hour) / 24.

        jd[i] = (int(365.25 * (year + 4716)) +
                 int(30.6001 * (month + 1)) + t.day + B - 1524.5 + C)

    return jd.squeeze()


def julian2sidereal(juliandate: float) -> float:
    """
    Convert Julian time to sidereal time

    D. Vallado Ed. 4

    input:

    juliandate
        Julian centuries from J2000.0

    """

    jdate = np.atleast_1d(juliandate)
    assert jdate.ndim == 1

    tsr = np.empty(jdate.size)
    for i, jd in enumerate(jdate):
        # %% Vallado Eq. 3-42 p. 184, Seidelmann 3.311-1
        tUT1 = (jd - 2451545.0) / 36525.

        # Eqn. 3-47 p. 188
        gmst_sec = (67310.54841 + (876600 * 3600 + 8640184.812866) *
                    tUT1 + 0.093104 * tUT1**2 - 6.2e-6 * tUT1**3)

        # 1/86400 and %(2*pi) implied by units of radians
        tsr[i] = gmst_sec * (2 * pi) / 86400. % (2 * pi)

    return tsr.squeeze()



#!/usr/bin/env python
import pytest
from pymap3d.timeconv import str2dt
from datetime import datetime

t0 = datetime(2014, 4, 6, 8)


def test_str2dt():
    assert str2dt(t0) == t0  # passthrough
    assert str2dt('2014-04-06T08:00:00') == t0
    ti = [str2dt('2014-04-06T08:00:00'), str2dt('2014-04-06T08:01:02')]
    to = [t0, datetime(2014, 4, 6, 8, 1, 2)]
    assert ti == to   # even though ti is numpy array of datetime and to is list of datetime


def test_xarray_time():
    xarray = pytest.importorskip('xarray')

    t = {'time': t0}

    ds = xarray.Dataset(t)
    assert str2dt(ds['time']) == t0


def test_pandas_time():
    pandas = pytest.importorskip('pandas')

    t = pandas.Series(t0)
    assert str2dt(t) == t0


if __name__ == '__main__':
    pytest.main([__file__])



#!/usr/bin/env python
import pytest
from pytest import approx
from pymap3d.vincenty import vreckon, vdist, track2
import numpy as np

az = [38., 45.]
sr = [3e3, 1e3]

lla0 = [42, -82, 200]
ll0 = [10, 20]

lla1 = [40, -80, 1120]


lat2 = [10.02137267, 10.01917819]
lon2 = [20.0168471, 20.0193493]
az2 = [218.00292856, 225.00336316]

lat3 = (10.02137267, 10.00639286)
lon3 = (20.0168471, 20.00644951)
az3 = (218.00292856, 225.0011203)


def test_track2():
    lats, lons = track2(40, 80, 65, -148, npts=3)
    assert lats == approx([40, 69.633139886, 65])
    assert lons == approx([80, 113.06849104, -148])


def test_vreckon():
    """ tests scalars, vectors"""
    pytest.importorskip('pytest', minversion='3.5')

    # scalar
    assert vreckon(*ll0, sr[0], az[0]) == approx((lat2[0], lon2[0], az2[0]))
    # az vector
    a, b, c = vreckon(*ll0, sr[0], az)
    assert a == approx(lat2)
    assert b == approx(lon2)
    assert c == approx(az2)
    # rng, az vectors
    a, b, c = vreckon(*ll0, sr, az)
    assert a == approx(lat3)
    assert b == approx(lon3)
    assert c == approx(az3)


def test_vdist():
    pytest.importorskip('pytest', minversion='3.5')

    lat1, lon1, a21 = vreckon(*ll0, sr[0], az[0])

    assert vdist(*ll0, lat1, lon1) == approx((sr[0], az[0], a21))
    # lat, lon vectors
    asr, aaz, aa21 = vdist(*ll0, lat2, lon2)

    assert np.all(sr[0] == approx(asr))  # for older pytest
    assert aaz == approx(az)


if __name__ == '__main__':
    pytest.main(['-xrsv', __file__])



#!/usr/bin/env python
import pytest
from pytest import approx
from pymap3d.vincenty import vreckon
import pymap3d as pm

lla0 = [42, -82, 200]


def test_compare_vicenty():
    taz, tsr = 38, 3000
    pyproj = pytest.importorskip('pyproj')

    lat2, lon2, a21 = vreckon(10, 20, tsr, taz)

    p4lon, p4lat, p4a21 = pyproj.Geod(ellps='WGS84').fwd(lon2, lat2, taz, tsr)
    assert (p4lon, p4lat, p4a21 % 360.) == approx((lon2, lat2, a21), rel=0.0025)

    p4az, p4a21, p4sr = pyproj.Geod(ellps='WGS84').inv(20, 10, lon2, lat2)
    assert (p4az, p4a21 % 360., p4sr) == approx((taz, a21, tsr))


def test_compare_geodetic():
    pyproj = pytest.importorskip('pyproj')

    xyz = pm.geodetic2ecef(*lla0)

    ecef = pyproj.Proj(proj='geocent', ellps='WGS84', datum='WGS84')
    lla = pyproj.Proj(proj='latlong', ellps='WGS84', datum='WGS84')

    assert pyproj.transform(lla, ecef, lla0[1], lla0[0], lla0[2]) == approx(xyz)
    assert pyproj.transform(ecef, lla, *xyz) == approx((lla0[1], lla0[0], lla0[2]))


if __name__ == '__main__':
    pytest.main(['-xrsv', __file__])



#!/usr/bin/env python
from pymap3d.haversine import anglesep, anglesep_meeus
from argparse import ArgumentParser
from pytest import approx


def main():
    p = ArgumentParser(description="angular distance between two sky points")
    p.add_argument('r0', help='right ascension: first point [deg]', type=float)
    p.add_argument('d0', help='declination: first point [deg]', type=float)
    p.add_argument('r1', help='right ascension: 2nd point [deg]', type=float)
    p.add_argument('d1', help='declination: 2nd point [degrees]', type=float)
    a = p.parse_args()

    dist_deg = anglesep_meeus(a.r0, a.d0, a.r1, a.d1)
    dist_deg_astropy = anglesep(a.r0, a.d0, a.r1, a.d1)

    print('{:.6f} deg sep'.format(dist_deg))

    assert dist_deg == approx(dist_deg_astropy)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# Copyright (c) 2014-2018 Michael Hirsch, Ph.D.
"""
Input/output: default units are METERS and DEGREES.
boolean deg=True means degrees

Most functions accept Numpy arrays of any shape
"""
from .timeconv import str2dt  # noqa: F401
from .azelradec import radec2azel, azel2radec  # noqa: F401
from .eci import eci2ecef, ecef2eci  # noqa: F401
from .datetime2hourangle import datetime2sidereal  # noqa: F401
from .ecef import geodetic2ecef, ecef2geodetic, eci2geodetic, Ellipsoid, ecef2enuv, enu2ecef, ecef2enu, enu2uvw, uvw2enu  # noqa: F401
from .ned import ned2ecef, ned2geodetic, geodetic2ned, ecef2nedv, ned2aer, aer2ned, ecef2ned  # noqa: F401
from .enu import enu2geodetic, geodetic2enu, aer2enu, enu2aer  # noqa: F401
from .aer import ecef2aer, aer2ecef, geodetic2aer, aer2geodetic, eci2aer, aer2eci  # noqa: F401
from .los import lookAtSpheroid  # noqa: F401
from .lox import isometric, meridian_dist, loxodrome_inverse  # noqa: F401



#!/usr/bin/env python
import pytest
from pytest import approx
import pymap3d as pm


def test_meridianarc():
    """
    meridianarc(0, deg2rad(40), wgs84Ellipsoid)
    """

    mdist = pm.meridian_dist(40)

    assert mdist == approx(4429529.03035058)


def test_rhumb():
    """
    distance('rh', 40, -80, 65, -148, wgs84Ellipsoid)
    azimuth('rh', 40, -80, 65, -148, wgs84Ellipsoid)
    """
    rhdist, rhaz = pm.loxodrome_inverse(40, -80, 65, -148)

    assert rhdist == approx(5248666.20853187)
    assert rhaz == approx(302.0056736)


if __name__ == '__main__':
    pytest.main(['-xrsv', __file__])



from typing import Tuple
from numpy import radians, sin, cos, hypot, arctan2, degrees
import numpy as np

from .ecef import geodetic2ecef, ecef2geodetic, enu2ecef, uvw2enu


def enu2aer(e: float, n: float, u: float, deg: bool=True) -> Tuple[float, float, float]:
    """
    ENU to Azimuth, Elevation, Range

    input
    -----
    e,n,u  [meters]  East, north, up                                [0,Infinity)
    deg    degrees input/output  (False: radians in/out)

    output: AER
    ------
    azimuth, elevation (degrees/radians)                             [0,360),[0,90]
    slant range [meters]                                             [0,Infinity)
    """
    r = hypot(e, n)
    slantRange = hypot(r, u)
    elev = arctan2(u, r)
    az = arctan2(e, n) % (2 * arctan2(0, -1))

    if deg:
        az = degrees(az)
        elev = degrees(elev)

    return az, elev, slantRange


def aer2enu(az: float, el: float, srange: float, deg: bool=True) -> Tuple[float, float, float]:
    """
    input:
    ------
    azimuth, elevation (degrees/radians)                             [0,360),[0,90]
    slant range [meters]                                             [0,Infinity)
    deg    degrees input/output  (False: radians in/out)

    output:
    -------
    e,n,u   East, North, Up [m]

    """
    if deg:
        el = radians(el)
        az = radians(az)

    with np.errstate(invalid='ignore'):
        if (np.asarray(srange) < 0).any():
            raise ValueError('Slant range \in  [0, Infinity)')

    r = srange * cos(el)

    return r * sin(az), r * cos(az), srange * sin(el)


def enu2geodetic(e: float, n: float, u: float,
                 lat0: float, lon0: float, h0: float,
                 ell=None, deg: bool=True) -> Tuple[float, float, float]:
    """

    input
    -----
    e,n,u   East, North, Up [m]
    Observer: lat0, lon0, h0 (altitude, meters)
    ell    reference ellipsoid
    deg    degrees input/output  (False: radians in/out)


    output:
    -------
    target: lat,lon, h  (degrees/radians,degrees/radians, meters)

    """

    x, y, z = enu2ecef(e, n, u, lat0, lon0, h0, ell, deg=deg)

    return ecef2geodetic(x, y, z, ell, deg=deg)


def geodetic2enu(lat: float, lon: float, h: float,
                 lat0: float, lon0: float, h0: float,
                 ell=None, deg: bool=True) -> Tuple[float, float, float]:
    """
    input
    -----
    target: lat,lon, h
    Observer: lat0, lon0, h0 (altitude, meters)
    ell    reference ellipsoid
    deg    degrees input/output  (False: radians in/out)


    output:
    -------
    e,n,u   East, North, Up [m]
    """
    x1, y1, z1 = geodetic2ecef(lat, lon, h, ell, deg=deg)
    x2, y2, z2 = geodetic2ecef(lat0, lon0, h0, ell, deg=deg)
    dx = x1 - x2
    dy = y1 - y2
    dz = z1 - z2

    return uvw2enu(dx, dy, dz, lat0, lon0, deg=deg)
